# Wrapper for phylo HMM
# Make sure your reference sequence is the first one of the fasta file.
# The reference sequence is the sequence which you are interested about.
# Filtering by domain and structure is performed on this reference sequence.
# A distance of 1 in the tree corresponds to ~20MY divergence. It significantly accelerates the procedure to scale your tree beforehand such that the closest species in your tree are within a distance of 1.
use threads;
use threads::shared;
use File::Basename;
my $dirname = dirname(__FILE__);
$SIG{INT} = \&interrupt;
if(scalar(@ARGV) < 4){
	help();
}
##################### Filepaths #################################

my $LR_exe = "$dirname/local_rate.exe";
my $phylo_HMM_perl = "$dirname/phylo_HMM.pl";
my $domain_perl = "$dirname/domain_hmm.pl";
#################################################################

my $cpu = 1;
my $quiet = 0;

my $input_file;
my $tree_file;

my $wag = "";

my $DOMAINS = 1;

my $debug = "";
my $threshold = 0.6; # Default threshold
my $peak_find = 0.2; # Default peak finding threshold
my $weight = 3; # Default conservation weight

my $c_distrib = "";
my $l_distrib = "";
my $length_alignment;

my $temp_1;
my $temp_2;

my $d_scale = 1;
for my $i (0 .. scalar(@ARGV)-1) {
	# Program files
	if ($ARGV[$i] eq "-cpu") { $cpu=$ARGV[$i+1]; }
	if ($ARGV[$i] eq "-quiet") { $quiet = 1; } # Do not print to screen progress.
	if ($ARGV[$i] eq "-wag") { $wag = "-wag"; } # Not recommended
	if ($ARGV[$i] eq "-distance") { $d_scale = $ARGV[$i+1];} # Do not tinker with this. In theory, setting a low -distance flag (less than 1) should improve accuracy at the expense of speed (and vice-versa). This has not been tested however.
	
	# Input files
	if ($ARGV[$i] eq "-i") { $input_file=$ARGV[$i+1]; }
	if ($ARGV[$i] eq "-tree") { $tree_file=$ARGV[$i+1]; }
	
	# Filtering steps
	# Use the flag to not perform the filtering step (you can filter the predicted motifs after if you've already pre-computed the regions).
	if ($ARGV[$i] eq "-domains") { $DOMAINS=0; }
	
	# Phylo-HMM parameters
	if ($ARGV[$i] eq "-debug") { $debug="-debug"; }
	if ($ARGV[$i] eq "-t") { $threshold = $ARGV[$i+1]; }
	if ($ARGV[$i] eq "-peak") { $peak_find = $ARGV[$i+1]; }
	if ($ARGV[$i] eq "-weight") { $weight = $ARGV[$i+1]; }
	
	# Precomputed files
	if( $ARGV[$i] eq "-loc_distrib"){ $l_distrib = $ARGV[$i+1]; }
	if( $ARGV[$i] eq "-col_distrib"){ $c_distrib = $ARGV[$i+1]; }

	# Help function
	if( $ARGV[$i] eq "-h" || $ARGV[$i] eq "-help"){ help(); }
}
my $count_complete = 0;
share($count_complete);
select(STDERR);
$| = 1;
select(STDOUT);
if(-e "$input_file" && -e "$tree_file"){
	$|=1;
	my $counter = 0;
	
	# Get length of alignment.
	open(INPUT_FILE,$input_file) or interrupt("file");
	my @sequences;
	my $seq_count = -1;
	while(my $lines = <INPUT_FILE>){
		chomp($lines);
		if($lines =~ m/>/){
			$seq_count += 1;
			$sequences[$seq_count]{"header"} = $lines;
		}
		else{
			$sequences[$seq_count]{"sequence"} .= $lines;	
		}
	}
	$length_alignment = length($sequences[0]{'sequence'});
	if($length_alignment < 1){ interrupt("file"); }
	my $ref_sequence = $sequences[0]{'sequence'};
	$ref_sequence =~ s/[^ACDEFGHIKLMNPQRSTVWY]//g;
	
	# Calculate the column rate of evolution.
	# Store column rate in temporary file.
	# If the column rate has been precomputed, skip
	my $spread_process = int($length_alignment / $cpu);
	
	if($c_distrib eq ""){
		my @threads;
		for(my $p = 1;$p < $length_alignment+1;$p=$p+$spread_process+1){
			if($cpu == 1){
				$threads[$p] = run_LR($p,min($p+$spread_process,$length_alignment),1);
			}
			else{
				$threads[$p] = threads->create(\&run_LR,$p,min($p+$spread_process,$length_alignment),1);		
			}
		}
		$temp_1 = getFreeFilename();
		open(COL_RATE_TEMP,">$temp_1") or interrupt("file_create");
		for(my ($p) = 1;$p < $length_alignment+1;$p=$p+$spread_process+1){
			my $print_list;
			if($cpu == 1){
				$print_list = $threads[$p];	
			}
			else{
				$print_list = $threads[$p]->join();
			}
			my @pl = split(/\n/, $print_list);
			for(my $i = 0;$i < scalar(@pl);++$i){
				my $line = $pl[$i];
				print COL_RATE_TEMP $line;
				print COL_RATE_TEMP "\n";
			}
		}
		close(COL_RATE_TEMP);
	}
	else{
		if(-e "$c_distrib"){
			$temp_1 = $c_distrib;
		}
		else{
			interrupt("c_rate");	
		}		
	}
	$count_complete = 0;
	if($l_distrib eq ""){
		# Calculate the local rate of evolution.
		# Store column rate in temporary file. 
		my @threads;
		for(my $p = 1;$p < $length_alignment+1;$p=$p+$spread_process+1){
			if($cpu == 1){
				$threads[$p] = run_LR($p,min($p+$spread_process,$length_alignment),21);
			}
			else{
				$threads[$p] = threads->create(\&run_LR,$p,min($p+$spread_process,$length_alignment),21);		
			}
		}
		$temp_2 = getFreeFilename();
		open(LOCAL_RATE_TEMP,">$temp_2") or interrupt("file_create");
		for(my ($p) = 1;$p < $length_alignment+1;$p=$p+$spread_process+1){
			my $print_list;
			if($cpu == 1){
				$print_list = $threads[$p];	
			}
			else{
				$print_list = $threads[$p]->join();
			}
			my @pl = split(/\n/, $print_list);
			for(my $i = 0;$i < scalar(@pl);++$i){
				my $line = $pl[$i];
				print LOCAL_RATE_TEMP $line;
				print LOCAL_RATE_TEMP "\n";
			}
		}
		close(LOCAL_RATE_TEMP);
	}
	else{
		if(-e "$l_distrib"){
			$temp_2 = $l_distrib;	
		}
		else{
			interrupt("l_rate");	
		}
	}
	
	# Predict domains
	my %domains;
	if($DOMAINS == 1){
		print STDERR "\r".(" " x 70)."\rPredicting large domains" unless ($quiet == 1);
		open(DOMAINS,"perl $domain_perl -i $input_file -tree $tree_file -distrib $temp_1 -l 100 $wag -min 20 -max 0 -cpu $cpu -t 0.6 -weight $weight |") or interrupt("domains");
		while(my $lines = <DOMAINS>){
			if($lines =~ m/([0-9]+)-([0-9]+)/){
				$domains{$1} = $2;	
			}
		}
		print STDERR "\r".(" " x 70)."\r" unless ($quiet == 1);
	}
	
	print STDERR "\r".(" " x 70)."\rPredicting" unless ($quiet == 1);
	open(PHYLO,"perl $phylo_HMM_perl -i $input_file -l 2 -min 2 -max 20 -cpu $cpu -t $threshold -peak $peak_find $debug $wag -w 10 -col_distrib $temp_1 -loc_distrib $temp_2 -tree $tree_file -weight $weight -distance $d_scale |") or interrupt("phylo_HMM");
	
	if($debug ne ""){
		$flag_debug = 1;	
	}
	my $flag_to_print = 0;
	my $site_count = 0;
	my $curr_start;
	my $curr_stop;
	my $tree_parsed;
	print STDERR "\r".(" " x 70)."\r" unless ($quiet == 1);
	while(my $lines = <PHYLO>){
		
		if($lines =~ m/Site\s#([0-9]+)\s.+?([0-9]+)-([0-9]+)(.*)/){
			chomp($lines);
			$flag_debug = 0;
			$flags = 0;
			$curr_start = $2;
			$curr_stop = $3;
			# Remove motifs predicted within these domains
			foreach my $starts (keys %domains){ # $domains{$start} = $stop;
				if(($2 >= $starts && $2 <= $domains{$starts})	|| ($3 >= $starts && $3 <= $domains{$starts})){
					$flags = 1;	
				}
			}
			
			if($flags == 1){
				$flag_to_print = 0;	
			}
			else{
				++$site_count;
				print "Site #$site_count : $2-$3$4\n";
				$flag_to_print = 1;
				next;
			}
		}
		if($lines =~ m/Tree\sused\s:\s(.+)/){
			$tree_parsed = $1;	
		}
		if($lines =~ m/(.+)(Found\s)[0-9]+(\ssites)/){
			print $1;
			print $2;
			print $site_count;
			print $3;
			print "\n";
			next;
		}
		if($flag_to_print == 1){
			print $lines;
		}
		if($flag_debug == 1){
			print $lines;	
		}
	}
	close(PHYLO);
}
else{
	interrupt("file");	
}
interrupt();
sub interrupt {
	my $error = @_[0];
	
	if($error eq "file"){
		print STDERR "\nCouldn't open or read properly the input files.\n";	
	}
	if($error eq "rate"){
		print STDERR "\nCouldn't calculate the local rate.\n";	
	}
	if($error eq "file_create"){
		print STDERR "\nCouldn't create the necessary temporary files.\n";	
	}
	if($error eq "domains"){
		print STDERR "\nCouldn't scan for domains.\n";	
	}
	if($error eq "success"){
		
		
	}
	if($error eq "phylo_HMM"){
		print STDERR "\nCouldn't run prediction program.\n";	
	}
	if($error eq "l_rate"){
		print STDERR "\nCouldn't find local rate precomputed file.\n";	
	}
	if($error eq "c_rate"){
		print STDERR "\nCouldn't find column rate precomputed file.\n";	
	}
	print STDERR "\nExiting script, cleaning up created files.\n";
	unlink($temp_1) unless (-e "$c_distrib" && $c_distrib ne "");
	unlink($temp_2) unless (-e "$l_distrib" && $l_distrib ne "");
	exit;
}
sub run_LR {
	my $start = @_[0];
	my $stop = @_[1];
	my $window = @_[2];
	my @print_array;
	my $command = "$LR_exe -i $input_file -tree $tree_file $wag -w $window -start $start -stop $stop -distance $d_scale |"; 
	my $which = "local";
	if($window == 1){
		$which = "column";	
	}
	open(CML,$command) or interrupt("rate");
	while(my $line = <CML>){
		push(@print_array,$line);
		lock($count_complete);
		++$count_complete;
		print STDERR "\r".(" " x 70)."\rCalculating $which rate of evolution\t".($count_complete/$length_alignment*100)."\%"  unless ($quiet == 1);
	}
	close(CML);
	return join("",@print_array);
}
sub getFreeFilename {
	while(1) {
		my $file_name = sprintf("%s_%s_%s",$$,$^T,$counter++);
		return $file_name if not -e $file_name;
	};
}
sub min {
	my($min) = shift(@_);

	foreach $temp (@_) {
		$min = $temp if $temp < $min;
	}
	return($min);
}
sub help {
	print STDERR "wrapper.pl -i file -tree file [-cpu int -loc_distrib file -col_distrib file -t float -weight float -debug -domains -quiet]\n";
	print STDERR "-i input_file (Required)\n";
	print STDERR "-tree tree_file (Required)\n";
	print STDERR "-cpu cpu_count [optional: specifies how many CPU to use during the prediction process.]\n";
	print STDERR "-loc_distrib local_distribution_file [optional: used when the local rate distribution has been precomputed]\n";
	print STDERR "-col_distrib column_distribution_file [optional: used when the column rate distribution has been precomputed]\n";
	print STDERR "-t posterior_threshold [default 0.6: specifies the probability threshold of prediction, must be between 0-1]\n";
	print STDERR "-weight conservation_weight [default 3: specifies the maximal fraction of conservation between the conserved state and the BG state. 3 sets to 1/3, 5 sets to 1/5...]\n";
	print STDERR "-domains [optional flag: Doing this flag will not filter the motifs by long conserved domain and return every peaks.]\n";
	print STDERR "-quiet [optional flag: will not print progress to screen]\n";
	print STDERR "-wag [optional flag: use the WAG matrix for globular proteins instead]\n";
	
	exit;
}